#  Spectrum::CLI - a perl module for use with Spectrum's Command Line Interface
#  Copyright (C) 1999-2000  Dave Plonka
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 2 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

# $Id: CLI.pm,v 1.11 2000/01/21 22:08:07 dplonka Exp $
# Dave Plonka <plonka@doit.wisc.edu>

package Spectrum::CLI;

use strict;
use vars qw($VERSION @ISA @EXPORT @EXPORT_OK);
use IO::File;
use IPC::Open2;
use File::Basename;

require Exporter;
require AutoLoader;

@Spectrum::CLI::ISA = qw(Exporter AutoLoader);
# convert the RCS revision to a reasonable Exporter VERSION:
'$Revision: 1.11 $' =~ m/(\d+)\.(\d+)/ && (( $Spectrum::CLI::VERSION ) = sprintf("%d.%03d", $1, $2));

sub new {
   my $class = shift;
   my $self = {};
   $self->{verbose} = 0;
   $ENV{CLISESSID} = $$; # default
   $ENV{CLIMNAMEWIDTH} = 1024; # default
   while (@_) {
      if ('HASH' eq ref($_[0])) { # a hashref
	 $self->{verbose} = $_[0]->{verbose} || $_[0]->{Verbose};
	 $self->{Verbose} = $_[0]->{Verbose};
	 if (defined $_[0]->{CLISESSID}) {
            $ENV{CLISESSID} = $_[0]->{CLISESSID}
	 }
	 if ($_[0]->{CLIMNAMEWIDTH}) {
            $ENV{CLIMNAMEWIDTH} = $_[0]->{CLIMNAMEWIDTH}
	 }
	 shift
      } else { # scalar
	 if (m|^/|) { # it's an absolute path
            $self->{dir} = shift
	 } else { # it's the VNM name
            $self->{vnm} = shift
	 }
      }
   }
   if ('' eq $self->{dir}) {
      if ($ENV{SPEC_ROOT}) {
         $self->{dir} = "$ENV{SPEC_ROOT}/vnmsh"
      } elsif ($ENV{SPECROOT}) {
         $self->{dir} = "$ENV{SPECROOT}/vnmsh"
      }
   }

   if ('' eq $self->{dir}) {
      warn "must set SPEC_ROOT/SPECROOT environment variable or pass full path to vnmsh directory to \"new\" method!\n" if $self->{verbose};
      return undef
   }

   my $command = "$self->{dir}/connect $self->{vnm}";
   my $fh = new IO::File "${command} 2>&1 |";
   if (!ref($fh)) {
      warn("failed to pipe from \"${command}\": $!\n") if $self->{verbose};
      return undef
   }
   @{$self->{results}} = <$fh>;
   $fh->close;
   $self->{status} = int($?/256);
   if (0 != $self->{status}) {
      warn("\"${command}\" failed - exit status: ", $self->{status}, "\n") if $self->{verbose};
      return undef
   }

   # discover which vnmsh commands are available:
   my $entry;
   my $dir = $self->{dir}; # kludge since perl doesn't like this in a glob
   foreach $entry (<$dir/*>) {
      next if $entry =~ m;/(Vnm|stop)Shd$;; # skip these executables
      push(@{$self->{command}}, basename $entry) if (-f $entry && -x _)

   }
   warn "valid vnmsh commands: @{$self->{command}}\n" if $self->{Verbose};

   bless($self, $class)
}

sub dir {
   my $self = shift;
   die unless ref($self);
   if (@_) {
      $self->{dir} = shift
   }
   $self->{dir}
}

sub verbose {
   my $self = shift;
   die unless ref($self);
   if (@_) {
      $self->{verbose} = shift
   } else {
      $self->{verbose}
   }
}

sub Verbose {
   my $self = shift;
   die unless ref($self);
   if (@_) {
      $self->{Verbose} = shift
   } else {
      $self->{Verbose}
   }
}

sub results {
   my $self = shift;
   die unless ref($self);
   @{$self->{results}}
}

sub status {
   my $self = shift;
   die unless ref($self);
   $self->{status}
}

sub AUTOLOAD {
   my $self = shift;
   my $call = $Spectrum::CLI::AUTOLOAD;
   $call =~ s/^.*:://;

   # { save the hashref, if one was passed, and remove it from the argument list
   my $hashref;
   my $n;
   for ($n=0; $n <= @_; $n++) {
      if ('HASH' eq ref($_[$n])) {
	 $hashref = splice(@_, $n, 1)
      }
   }
   # }

   my @command = (split(m/_/, $call), @_);

   if (!grep { $command[0] eq $_ } @{$self->{command}}) {
      die "Can't locate object method \"$call\""
   }

   warn "@command...\n" if $self->{Verbose};

   my $display = 1; # default to "display" commands such as "show" and "seek"
   if ('fetchall_arrayref' eq $command[0]) {
      $command[0] = 'show'
   } elsif ('show' ne $command[0] && 'seek' ne $command[0]) {
      $display = 0
   }

   my $fhin = new IO::File;
   die "IO::File->new failed: $!\n" unless ref($fhin);
   my $fh = new IO::File;
   die "IO::File->new failed: $!\n" unless ref($fh);
   my $pid = open2($fh, $fhin, $self->dir . "/@command 2>&1");
   print $fhin "y\n"; # answer 'y'es to "destroy model: are you sure ?", etc.

   if (!$display) {
      @{$self->{results}} = <$fh>;
      waitpid($pid, 0);
      $self->{status} = int($?/256);
      if (0 != $self->{status} && $self->{verbose}) {
	 warn @{$self->{results}}
      }
      $fh->close;
      $fhin->close;
      return wantarray? ($self->{status}, @{$self->{results}}) : !$self->{status}
   }

   # { parse the output of display commands:
   my @results;
   my $headings = <$fh>;
   @{$self->{results}} = ($headings);
   chomp $headings;
   my(@headings, @lengths);
   while ($headings) {
      $headings =~ s/^\s*(\S+)\s*//;
      push(@headings, $1);
      push(@lengths, length($&))
   }
   while (<$fh>) {
      push(@{$self->{results}}, $_);
      chomp;
      my $index = 1+$#results;
      my $start = 0; # position where this column starts
      my $lcv;
      for ($lcv = 0; $headings[$lcv]; $lcv++) {
	 my $val;
	 if ($lcv < $#headings) {
            $val = substr($_, $start, $lengths[$lcv]);
	    $start += $lengths[$lcv]
	 } else {
            $val = substr($_, $start)
	 }
	 if ($hashref) {
	    # skip columns that the caller didn't request...
	    next unless $hashref->{$headings[$lcv]}
	 }
	 $val =~ s/\s+$//;
         $results[$index]{$headings[$lcv]} = $val
      }
   }
   # }
   $fh->close;
   if (0 != $self->{status} && $self->{verbose}) {
      warn @{$self->{results}}
   }

   wantarray? @results : \@results
}

sub DESTROY {
   my $self = shift;
   die unless ref $self;
   system($self->dir . "/disconnect >/dev/null 2>&1");
}

1;
__END__

=head1 NAME

Spectrum::CLI - wrapper class for the Spectrum Command Line Interface

=head1 SYNOPSIS

  use Spectrum::CLI;

  my $cli = new Spectrum::CLI;

  # show_types:
  map { $Type{$_->{Name}} = $_ } $cli->show_types;

  # show_attributes:
  map {
     $UserAttr{$_->{Name}} = $_
  } $cli->show_attributes("mth=$Type{User}{Handle}");

  # create_model:
  $cli->create_model("mth=$Type{User}{Handle}",
                     "attr=$UserAttr{Model_Name}{Id},val=joe",
                     "attr=$UserAttr{User_Full_Name}{Id},val='Joe User'");

=head1 DESCRIPTION

C<Spectrum::CLI> is class which intends to provide a convenient way to
invoke the various CLI commands of Spectrum Enterprise Manager from
within perl scripts.

In the author's estimation, the two primary benefits of
C<Spectrum::CLI> are:

=over 4

=item the parsing of "show" command results

C<Spectrum::CLI> intelligently parses of the output of CLI "show"
commands.  That is, it will split apart the columnar values for you,
and return them as an array of hashes, each element representing one
line of output.

=item the elimination of "SpecHex" numbers in scripts

Because of the aforementioned results parsing, it is now easy to
mention spectrum data objects by name rather than by their hexadecimal
Handle or Id.  For instance, the following one=liner will create a hash
of models by model name:

    map { $Model{$_->{MName} = $_ } } $cli->show_models;

This would subsequently enable one to refer to a model's handle
thusly:

    my $handle = $Model{Universe}{MHandle};

In this way, scripts can be written which discover all SpecHex magic
numbers, and therefore are potentially more portable/reusable amongst
spectrum installations.

=back

=head1 METHODS

=over 4

=item new ( [VNM_Name] [,SPECROOT] [,options_hashref] )

This is the class' constructor - it returns a C<Spectrum::CLI> object
upon success or undef on failure.  Its arguments are optional, and may
occur in any order.  In general, it can be called with no arguments
unless the SPECROOT environment variable is not set, or one wishes to
connect to a SpectroSERVER on another host (i.e. the script is not
running on the VNM host).

The optional constructor arguments are:

=over 4

=item VNM name

You may pass a string specifying the name of the VNM machine where the
SpectroSERVER to which you'd like to connect resides.  (This is passed
to the as the first argument to the "connect" command.)

=item SPECROOT value

You may pass a string specifying the absolute path to the "vnmsh"
directory.  It should point to a directory which contains the "vnmsh"
sub-directory.  (It is not necessary to pass this argument if either
the SPECROOT or SPEC_ROOT environment variables are set.)

=item options hashref

A hash reference may also be passed as an argument.  The hash can be
used to specify a number of options.  These include:

=over 4

=item verbose

Passing { verbose => 1 } will cause the package to report errors from
the CLI commands to STDERR.  This feature can also be enabled by
calling the "verbose" object method with a non-zero argument after the
constructor returns successfully with that object.

=item Verbose

Passing { Verbose => 1 } will cause the package to report what it is
doing, such as which CLI command it is about to execute.  This feature
can also be enabled by calling the "Verbose" object method with a
non-zero argument after the constructor returns successfully with that
object.

=item CLISESSID

Passing { CLISESSID => value } will cause C<Spectrum::CLI> to use that
value as the session id.  Normally, in the absence of this option, the
perl script's process id is used.  Use of this option is discouraged
unless the default values is somehow problematic.

=item CLIMNAMEWIDTH

Passing { CLIMNAMEWIDTH => value } will cause C<Spectrum::CLI> to use
that value as the model name width for CLI "show" commands. Normally,
in the absence of this option, the value 1024 is used.  Use of this
option is discouraged unless the default values is somehow
problematic.

=back

=back

=item show_* ( [ARGS] )

These methods invoke the CLI "show" command.  They return an array of
hashes where the hash keys are the column headings from the first line
of output of the corresponding CLI command, and the values are the
corresponding value in that column for the given row.

=item seek (attr=attribute_id,val=value [,lh=landscape_handle])

This method invokes the CLI "seek" command.  It returns values in a
like the "show" methods.

=item create_*,destroy_*,etc. ( ARGS [,...] )

ALL of the other CLI commands are available as methods.  The return
value(s) of these methods differs markedly from that of the
aforementioned show_* or seek methods.  This is because these methods
do not normally produce columnar output, but instead produce messages
which sometimes include some useful piece of information.  For
instance, "create model" produces a message upon success, which
indicates the model handle of the create model.  The caller will nearly
always want to obtain that handle, which can be done by parsing the
returned values.  (See the C<results> method for a hint at how to do
this.)

In an array context these methods return an array in which the first
element is the exit status (i.e. zero indication success) of the
underlying CLI command.  The subsequent elements of the returned array
are the standard output and standard error (if any) lines produced by
the underlying CLI command.

In a scalar context these methods return non-zero if the command
succeeds, zero otherwise.

Regardless of the context in which they are invoked, these methods
cause the objects internal status and results to be set.  If it is more
convenient, these values can be retrieved using the C<status> and
C<results> methods rather than having to collect them as return
values.

=item dir

This method returns the absolute path to the directory containing the
vnmsh commands.

=item verbose ([ARG])

This method returns a boolean value indicating whether or not
verbose behavior is currently turned on.

If a zero or non-zero argument is passed it will either clear or set
this feature, respectively. 

=item Verbose ([ARG])

This method returns a boolean value indicating whether or not "very
verbose" behavior is currently turned on.

If a zero or non-zero argument is passed it will either clear or set
this feature, respectively.

=item results

This method returns an array containing the results (standard output
and standard error) of the last vnmsh command that was executed.

It is *extremely* useful to determine things such as the model handle
of the model that was created by a call the "create_model".

=item status

This method returns the exit status of the last vnmsh command that was
executed.

=back

=head1 AUTHOR

Dave Plonka <plonka@doit.wisc.edu>

Copyright (C) 1999-2000  Dave Plonka.
This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

=head1 SEE ALSO

perl(1).

=cut
